import json
import re
from typing import TypedDict

from aiohttp import ClientSession

from yutto.api.info import get_video_info, VideoInfo, PageInfo
from yutto.exceptions import NoAccessPermissionError, UnSupportedTypeError
from yutto.bilibili_typing.codec import audio_codec_map, video_codec_map
from yutto._typing import AudioUrlMeta, AvId, CId, MultiLangSubtitle, VideoUrlMeta
from yutto.utils.console.logger import Logger
from yutto.utils.fetcher import Fetcher
from yutto.utils.metadata import MetaData
from yutto.utils.time import get_time_str_by_now, get_time_str_by_stamp


class AcgVideoListItem(TypedDict):
    id: int
    name: str
    avid: AvId
    cid: CId
    metadata: MetaData


class AcgVideoList(TypedDict):
    title: str
    pubdate: str
    pages: list[AcgVideoListItem]


async def get_acg_video_list(session: ClientSession, avid: AvId) -> AcgVideoList:
    video_info = await get_video_info(session, avid)
    result: AcgVideoList = {
        "title": video_info["title"],
        "pubdate": get_time_str_by_stamp(video_info["pubdate"], "%Y-%m-%d"),  # TODO: 可自由定制
        "pages": [],
    }
    list_api = "https://api.bilibili.com/x/player/pagelist?aid={aid}&bvid={bvid}&jsonp=jsonp"
    res_json = await Fetcher.fetch_json(session, list_api.format(**avid.to_dict()))
    if res_json is None or res_json.get("data") is None:
        Logger.warning(f"啊叻？视频 {avid} 不见了诶")
        return result
    result["pages"] = [
        {
            "id": i + 1,
            "name": item["part"],
            "avid": avid,
            "cid": CId(str(item["cid"])),
            "metadata": _parse_acg_video_metadata(video_info, page_info),
        }
        for i, (item, page_info) in enumerate(zip(res_json["data"], video_info["pages"]))
    ]
    return result


async def get_acg_video_playurl(
    session: ClientSession, avid: AvId, cid: CId
) -> tuple[list[VideoUrlMeta], list[AudioUrlMeta]]:
    # 4048 = 16(useDash) | 64(useHDR) | 128(use4K) | 256(useDolby) | 512(useXXX) | 1024(use8K) | 2048(useAV1)
    play_api = "https://api.bilibili.com/x/player/playurl?avid={aid}&bvid={bvid}&cid={cid}&qn=127&type=&otype=json&fnver=0&fnval=4048&fourk=1"

    resp_json = await Fetcher.fetch_json(session, play_api.format(**avid.to_dict(), cid=cid))
    if resp_json is None:
        raise NoAccessPermissionError(f"无法获取该视频链接（avid: {avid}, cid: {cid}）")
    if resp_json.get("data") is None:
        raise NoAccessPermissionError(f"无法获取该视频链接（avid: {avid}, cid: {cid}），原因：{resp_json.get('message')}")
    if resp_json["data"].get("dash") is None:
        raise UnSupportedTypeError(f"该视频（avid: {avid}, cid: {cid}）尚不支持 DASH 格式")
    # TODO: 处理 resp_json["data"]["dash"]["dolby"]，应当是 Dolby 的音频流
    return (
        [
            {
                "url": video["base_url"],
                "mirrors": video["backup_url"] if video["backup_url"] is not None else [],
                "codec": video_codec_map[video["codecid"]],
                "width": video["width"],
                "height": video["height"],
                "quality": video["id"],
            }
            for video in resp_json["data"]["dash"]["video"]
        ]
        if resp_json["data"]["dash"]["video"]
        else [],
        [
            {
                "url": audio["base_url"],
                "mirrors": audio["backup_url"] if audio["backup_url"] is not None else [],
                "codec": audio_codec_map[audio["codecid"]],
                "width": 0,
                "height": 0,
                "quality": audio["id"],
            }
            for audio in resp_json["data"]["dash"]["audio"]
        ]
        if resp_json["data"]["dash"]["audio"]
        else [],
    )


async def get_acg_video_subtitles(session: ClientSession, avid: AvId, cid: CId) -> list[MultiLangSubtitle]:
    subtitile_api = "https://api.bilibili.com/x/player.so?aid={aid}&bvid={bvid}&id=cid:{cid}"
    subtitile_url = subtitile_api.format(**avid.to_dict(), cid=cid)
    res_text = await Fetcher.fetch_text(session, subtitile_url)
    if res_text is None:
        return []
    if subtitle_json_text_match := re.search(r"<subtitle>(.+)</subtitle>", res_text):
        subtitle_json = json.loads(subtitle_json_text_match.group(1))
        results: list[MultiLangSubtitle] = []
        for sub_info in subtitle_json["subtitles"]:
            subtitle_text = await Fetcher.fetch_json(session, "https:" + sub_info["subtitle_url"])
            if subtitle_text is None:
                continue
            results.append(
                {
                    "lang": sub_info["lan_doc"],
                    "lines": subtitle_text["body"],
                }
            )
        return results
    return []


def _parse_acg_video_metadata(video_info: VideoInfo, page_info: PageInfo) -> MetaData:
    return MetaData(
        title=page_info["part"],
        show_title=page_info["part"],
        plot=video_info["description"],
        thumb=page_info["first_frame"] if page_info["first_frame"] is not None else video_info["picture"],
        premiered=get_time_str_by_stamp(video_info["pubdate"]),
        dataadded=get_time_str_by_now(),
        source="",  # TODO
        original_filename="",  # TODO
    )



import heapq
import os
from dataclasses import dataclass, field
from types import TracebackType
from typing import Optional, Type

import aiofiles

from yutto.utils.console.logger import Logger
from yutto.utils.functools import aobject


@dataclass(order=True)
class BufferChunk:
    offset: int
    data: bytes = field(compare=False)


class AsyncFileBuffer(aobject):
    """异步文件缓冲区

    Args:
        file_path (str): 所需存储文件位置
        overwrite (bool): 是否直接覆盖原文件

    Examples:
        .. code-block:: python

            async def afunc():
                buffer = await AsyncFileBuffer("/path/to/file", True)
                for i, chunk in enumerate([b'0', b'1', b'2', b'3', b'4']):
                    await buffer.write(chunk, i)
                await buffer.close()

                # 或者使用 async with（注意后面要有 await，因为 AsyncFileBuffer 的初始化是异步的）

                async with await AsyncFileBuffer("/path/to/file", True) as buffer:
                    for i, chunk in enumerate([b'0', b'1', b'2', b'3', b'4']):
                        await buffer.write(chunk, i)
    """

    # pyright: reportIncompatibleMethodOverride=false
    async def __ainit__(self, file_path: str, overwrite: bool = False):
        self.file_path = file_path
        if overwrite and os.path.exists(file_path):
            os.remove(file_path)
        self.buffer = list[BufferChunk]()
        self.written_size = os.path.getsize(file_path) if not overwrite and os.path.exists(file_path) else 0
        self.file_obj: Optional[aiofiles.threadpool.binary.AsyncBufferedIOBase] = await aiofiles.open(file_path, "ab")

    async def write(self, chunk: bytes, offset: int):
        buffer_chunk = BufferChunk(offset, chunk)
        # 使用堆结构，保证第一个元素始终最小
        heapq.heappush(self.buffer, buffer_chunk)
        while self.buffer and self.buffer[0].offset <= self.written_size:
            assert self.file_obj is not None
            ready_to_write_chunk = heapq.heappop(self.buffer)
            if ready_to_write_chunk.offset < self.written_size:
                Logger.error(f"交叠的块范围 {ready_to_write_chunk.offset} < {self.written_size}，舍弃！")
                continue
            await self.file_obj.write(ready_to_write_chunk.data)
            self.written_size += len(ready_to_write_chunk.data)

    async def close(self):
        if self.buffer:
            Logger.error("buffer 尚未清空")
        if self.file_obj is not None:
            await self.file_obj.close()
        else:
            Logger.error("未预期的结果：未曾创建文件对象")

    def __enter__(self) -> None:
        raise TypeError("Use async with instead")

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        # __exit__ should exist in pair with __enter__ but never executed
        ...

    async def __aenter__(self) -> "AsyncFileBuffer":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        await self.close()



import asyncio
import platform

from yutto.utils.console.logger import Logger


def initial_async_policy():
    if platform.system() == "Windows":
        Logger.debug("Windows 平台，单独设置 EventLoopPolicy")
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore


def install_uvloop():
    try:
        import uvloop  # type: ignore
    except ImportError:
        Logger.warning("未安装 uvloop，无法使用其加速协程")
    else:
        uvloop.install()  # type: ignore
        Logger.info("成功使用 uvloop 加速协程")



import re
from typing import Optional, TypedDict

from aiohttp import ClientSession

from yutto._typing import AId, AvId, BvId, CId, EpisodeId
from yutto.exceptions import NotFoundError
from yutto.utils.fetcher import Fetcher


class PageInfo(TypedDict):
    part: str
    first_frame: Optional[str]


class VideoInfo(TypedDict):
    avid: AvId
    aid: AId
    bvid: BvId
    episode_id: EpisodeId
    is_bangumi: bool
    cid: CId
    picture: str
    title: str
    pubdate: int
    description: str
    pages: list[PageInfo]


async def get_video_info(session: ClientSession, avid: AvId) -> VideoInfo:
    regex_ep = re.compile(r"https?://www\.bilibili\.com/bangumi/play/ep(?P<episode_id>\d+)")
    info_api = "http://api.bilibili.com/x/web-interface/view?aid={aid}&bvid={bvid}"
    res_json = await Fetcher.fetch_json(session, info_api.format(**avid.to_dict()))
    if res_json is None:
        raise NotFoundError(f"无法该视频 {avid} 信息")
    res_json_data = res_json.get("data")
    if res_json["code"] == 62002:
        raise NotFoundError(f"无法下载该视频 {avid}，原因：{res_json['message']}")
    if res_json["code"] == -404:
        raise NotFoundError(f"啊叻？视频 {avid} 不见了诶")
    assert res_json_data is not None, "响应数据无 data 域"
    episode_id = EpisodeId("")
    if res_json_data.get("redirect_url") and (ep_match := regex_ep.match(res_json_data["redirect_url"])):
        episode_id = EpisodeId(ep_match.group("episode_id"))
    return {
        "avid": BvId(res_json_data["bvid"]),
        "aid": AId(str(res_json_data["aid"])),
        "bvid": BvId(res_json_data["bvid"]),
        "episode_id": episode_id,
        "is_bangumi": bool(episode_id),
        "cid": CId(str(res_json_data["cid"])),
        "picture": res_json_data["pic"],
        "title": res_json_data["title"],
        "pubdate": res_json_data["pubdate"],
        "description": res_json_data["desc"],
        "pages": [
            {
                "part": page["part"],
                "first_frame": page.get("first_frame"),
            }
            for page in res_json_data["pages"]
        ],
    }


async def is_vip(session: ClientSession) -> bool:
    info_api = "https://api.bilibili.com/x/web-interface/nav"
    res_json = await Fetcher.fetch_json(session, info_api)
    assert res_json is not None
    res_json_data = res_json.get("data")
    if res_json_data.get("vipStatus") == 1:
        return True
    return False



import argparse
import asyncio
import re
from typing import Any, Coroutine, Optional

import aiohttp

from yutto._typing import EpisodeData, MId, SeriesId
from yutto.api.acg_video import AcgVideoListItem, get_acg_video_list
from yutto.api.space import (
    get_collection_avids,
    get_collection_title,
    get_medialist_avids,
    get_medialist_title,
    get_uploader_name,
)
from yutto.exceptions import NotFoundError
from yutto.extractor._abc import BatchExtractor
from yutto.extractor.common import extract_acg_video_data
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.fetcher import Fetcher


class SeriesExtractor(BatchExtractor):
    """视频合集和视频列表"""

    REGEX_SERIES = re.compile(
        r"https?://space\.bilibili\.com/(?P<mid>\d+)/channel/seriesdetail\?sid=(?P<series_id>\d+)"
    )
    REGEX_MEDIA_LIST = re.compile(
        r"https?://www\.bilibili\.com/medialist/play/(?P<mid>\d+)\?business=space_series&business_id=(?P<series_id>\d+)"
    )
    REGEX_COLLECTIOMS = re.compile(
        r"https?://space\.bilibili\.com/(?P<mid>\d+)/channel/collectiondetail\?sid=(?P<series_id>\d+)"
    )

    mid: MId
    series_id: SeriesId
    is_collection: bool

    def match(self, url: str) -> bool:
        if (
            (match_obj := self.REGEX_MEDIA_LIST.match(url))
            or (match_obj := self.REGEX_SERIES.match(url))
            or (match_obj := self.REGEX_COLLECTIOMS.match(url))
        ):
            self.mid = MId(match_obj.group("mid"))
            self.series_id = SeriesId(match_obj.group("series_id"))
            self.is_collection = True if self.REGEX_COLLECTIOMS.match(url) else False
            return True
        else:
            return False

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Optional[Coroutine[Any, Any, Optional[EpisodeData]]]]:
        # 视频合集
        if self.is_collection:
            username, series_title = await asyncio.gather(
                get_uploader_name(session, self.mid),
                get_collection_title(session, self.series_id),
            )
            Logger.custom(series_title, Badge("视频合集", fore="black", back="cyan"))

            acg_video_info_list: list[tuple[AcgVideoListItem, str, str]] = []
            for avid in await get_collection_avids(session, self.series_id):
                try:
                    acg_video_list = await get_acg_video_list(session, avid)
                    await Fetcher.touch_url(session, avid.to_url())
                    for acg_video_item in acg_video_list["pages"]:
                        acg_video_info_list.append(
                            (
                                acg_video_item,
                                acg_video_list["title"],
                                acg_video_list["pubdate"],
                            )
                        )
                except NotFoundError as e:
                    Logger.error(e.message)
                    continue
        # 视频列表
        else:
            username, series_title = await asyncio.gather(
                get_uploader_name(session, self.mid), get_medialist_title(session, self.series_id)
            )
            Logger.custom(series_title, Badge("视频列表", fore="black", back="cyan"))

            acg_video_info_list: list[tuple[AcgVideoListItem, str, str]] = []
            for avid in await get_medialist_avids(session, self.series_id):
                try:
                    acg_video_list = await get_acg_video_list(session, avid)
                    await Fetcher.touch_url(session, avid.to_url())
                    for acg_video_item in acg_video_list["pages"]:
                        acg_video_info_list.append(
                            (
                                acg_video_item,
                                acg_video_list["title"],
                                acg_video_list["pubdate"],
                            )
                        )
                except NotFoundError as e:
                    Logger.error(e.message)
                    continue

        return [
            extract_acg_video_data(
                session,
                acg_video_item["avid"],
                acg_video_item,
                args,
                {
                    "series_title": series_title,
                    "username": username,  # 虽然默认模板的用不上，但这里可以提供一下
                    "title": title,
                    "pubdate": pubdate,
                },
                "{series_title}/{title}/{name}",
            )
            for acg_video_item, title, pubdate in acg_video_info_list
        ]



import argparse
from typing import Any, Coroutine, Optional, TypeVar

import aiohttp

from yutto._typing import EpisodeData

T = TypeVar("T")


class Extractor:
    def resolve_shortcut(self, id: str) -> tuple[bool, str]:
        matched = False
        url = id
        return (matched, url)

    def match(self, url: str) -> bool:
        raise NotImplementedError

    async def __call__(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Optional[Coroutine[Any, Any, Optional[EpisodeData]]]]:
        raise NotImplementedError


class SingleExtractor(Extractor):
    async def __call__(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Optional[Coroutine[Any, Any, Optional[EpisodeData]]]]:
        return [await self.extract(session, args)]

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> Optional[Coroutine[Any, Any, Optional[EpisodeData]]]:
        raise NotImplementedError


class BatchExtractor(Extractor):
    async def __call__(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Optional[Coroutine[Any, Any, Optional[EpisodeData]]]]:
        return await self.extract(session, args)

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Optional[Coroutine[Any, Any, Optional[EpisodeData]]]]:
        raise NotImplementedError



import argparse
import re
from typing import Any, Coroutine, Optional

import aiohttp

from yutto._typing import AId, AvId, BvId, EpisodeData
from yutto.api.acg_video import get_acg_video_list
from yutto.exceptions import NotFoundError
from yutto.extractor._abc import BatchExtractor
from yutto.extractor.common import extract_acg_video_data
from yutto.processor.selector import parse_episodes_selection
from yutto.utils.console.logger import Badge, Logger


class AcgVideoBatchExtractor(BatchExtractor):
    """投稿视频批下载"""

    REGEX_AV = re.compile(r"https?://www\.bilibili\.com/video/av(?P<aid>\d+)(\?p=(?P<page>\d+))?")
    REGEX_BV = re.compile(r"https?://www\.bilibili\.com/video/(?P<bvid>(bv|BV)\w+)(\?p=(?P<page>\d+))?")

    REGEX_AV_ID = re.compile(r"av(?P<aid>\d+)(\?p=(?P<page>\d+))?")
    REGEX_BV_ID = re.compile(r"(?P<bvid>(bv|BV)\w+)(\?p=(?P<page>\d+))?")

    avid: AvId

    def resolve_shortcut(self, id: str) -> tuple[bool, str]:
        matched = False
        url = id
        if match_obj := self.REGEX_AV_ID.match(id):
            page: int = 1
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            url = f"https://www.bilibili.com/video/av{match_obj.group('aid')}?p={page}"
            matched = True
        elif match_obj := self.REGEX_BV_ID.match(id):
            page: int = 1
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            url = f"https://www.bilibili.com/video/{match_obj.group('bvid')}?p={page}"
            matched = True
        return matched, url

    def match(self, url: str) -> bool:
        if (match_obj := self.REGEX_AV.match(url)) or (match_obj := self.REGEX_BV.match(url)):
            if "aid" in match_obj.groupdict().keys():
                self.avid = AId(match_obj.group("aid"))
            else:
                self.avid = BvId(match_obj.group("bvid"))
            return True
        else:
            return False

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Optional[Coroutine[Any, Any, Optional[EpisodeData]]]]:
        try:
            acg_video_list = await get_acg_video_list(session, self.avid)
            Logger.custom(acg_video_list["title"], Badge("投稿视频", fore="black", back="cyan"))
        except NotFoundError as e:
            # 由于获取 info 时候也会因为视频不存在而报错，因此这里需要捕捉下
            Logger.error(e.message)
            return []

        # 选集过滤
        episodes = parse_episodes_selection(args.episodes, len(acg_video_list["pages"]))
        acg_video_list["pages"] = list(filter(lambda item: item["id"] in episodes, acg_video_list["pages"]))

        return [
            extract_acg_video_data(
                session,
                acg_video_item["avid"],
                acg_video_item,
                args,
                {
                    "title": acg_video_list["title"],
                    "pubdate": acg_video_list["pubdate"],
                },
                "{title}/{name}",
            )
            for acg_video_item in acg_video_list["pages"]
        ]



from typing import Any


class aobject(object):
    """Inheriting this class allows you to define an async __ainit__.

    Refs:
        https://stackoverflow.com/questions/33128325/how-to-set-class-attribute-with-await-in-init

    Examples:
        .. code-block:: python

            class MyClass(aobject):
                # pyright: reportIncompatibleMethodOverride=false
                async def __ainit__(self):
                    ...

            await MyClass()
    """

    async def __new__(cls, *args: Any, **kwargs: Any):
        instance = super().__new__(cls)
        await instance.__ainit__(*args, **kwargs)
        return instance

    async def __ainit__(self, *args: Any, **kwargs: Any):
        pass
